#! /usr/bin/python3
# -*- coding: utf-8 -*-
# @FileName: template.py
# @Author: Li Chengxin 
# @Time: 2023/6/30 22:26


REGIONS_TMP = {
    "shape_attributes":
        {
            "name": "polygon",
            "all_points_x": [],
            "all_points_y": []
        },
    "region_attributes":
        {
            "cell_type": None,
            "cell_id": None,
            'track_id': None,
        }
}

FRAME_TMP =  {
        "filename": None,
        "size": None,
        "regions": [],
        "file_attributes": {}
    }
